/*
** Module   : LargeObjectOps.java
** Abstract : helper routines to process LOBs
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20130516 First version implementing an early version of the COPY-LOB statement which
**                  is suitable for executing testcases but is not yet 100% complete.
** 002 EVK 20140206 Added method readFromFile. This is an early version of COPY-LOB statement
**                  which is suitable for executing testcases but is not yet 100% complete.
** 003 VIG 20140914 Added comment about behaviour in readFromFile method.
** 004 GES 20141002 Added simple implementation of copying between MEMPTR and LONGCHAR (either
**                  direction).  Also added alternate versions of readFromFile() that handle
**                  starting positions + length options.  All of these are only useful for simple
**                  testing, they are not 100% complete/compatible.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * Provides static utility methods for processing large objects (LOBs) in a Progress 4GL
 * compatible manner.
 * <p> 
 * At this time, the following Progress 4GL features are handled:
 * <p>
 * <pre>
 * Progress        From           To             Java Equivalent
 * -------------   ------------   ------------   ---------------------
 * COPY-LOB        MEMPTR         LONGCHAR       copy
 * COPY-LOB        LONGCHAR       MEMPTR         copy
 * COPY-LOB        File           MEMPTR         readFromFile
 * COPY-LOB        MEMPTR         File           writeToFile
 * </pre>
 */
public class LargeObjectOps
{
   /**
    * Private constructor to ensure instances are not created. All functionality of this class
    * is to be used via static methods.
    */
   private LargeObjectOps()
   {
   }
   
   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(BinaryData src, longchar target)
   {
      // TODO: make this fully compatible, this is just a partially working stub right now
      
      // unknown/uninitialized handling should be here
      
      byte[] contents = src.getByteArray();
      
      if (raiseErrorOnNullChar(contents))
         return;
      
      // code-page conversion needs to happen here
      target.assign(new String(contents));
   }

   /**
    * Copy the contents of one LOB to another.
    * <p>
    * No codepage conversion is implemented.  The entire contents will be copied.
    *
    * @param    src
    *           The object to copy from.
    * @param    target
    *           The object to copy to.
    */
   public static void copy(longchar src, BinaryData target)
   {
      // TODO: make this fully compatible, this is just a partially working stub right now
      
      // unknown handling should be here
      
      String val = src.toStringMessage();
      
      // code-page conversion needs to happen here
      byte[] contents = val.getBytes();
      
      target.assign(contents);
   }

   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, int len, BinaryData data)
   {
      // TODO: this is a simple placeholder that really should be replaced with a
      //       single call to a client-side worker that can handle all features, but
      //       until the feature set is fully explored we are implementing the simple
      //       way
      
      if (fname == null || fname.length() == 0 || data == null || data.isUnknown())
         return;
      
      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         // seek to the 0-based offset
         file.setPos(idx - 1);
         
         // read the data
         byte[] chunk = file.readChunk(len);
         
         // write into the buffer
         data.assign(chunk);
      }
      catch (EndConditionException | ErrorConditionException | IOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, int len, BinaryData data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, int idx, NumberType len, BinaryData data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, NumberType idx, NumberType len, BinaryData data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, int len, BinaryData data)
   {
      if (fname == null || fname.isUnknown())
         return;
      
      readFromFile(fname.toStringMessage(), idx, len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, NumberType idx, int len, BinaryData data)
   {
      if (idx == null || idx.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len, data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, int idx, NumberType len, BinaryData data)
   {
      if (len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx, len.intValue(), data);
   }
   
   /**
    * Read some contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    idx
    *           The 1-based offset into the file from which to start reading.
    * @param    len
    *           The number of bytes to read.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character  fname,
                                   NumberType idx,
                                   NumberType len,
                                   BinaryData data)
   {
      if (idx == null || idx.isUnknown() || len == null || len.isUnknown())
         return;
      
      readFromFile(fname, idx.intValue(), len.intValue(), data);
   }
   
   /**
    * Read the full contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(String fname, BinaryData data)
   {
// TODO: this is just a first pass, the real/full behavior should be implemented as following:
//      1. size(data) < size(file). For exapmle size(data) = 5 and size(file) = 23.
//      Progress reads all file and dynamically increases data size to needed value 
//      i.e. from 5 to 23 (in case of data).
//      Java impl only reads size(data) bytes from a file, i.e. 5
//      2. size(data) > size(file). For exapmle size(data) = 35 and size(file) = 23.
//      Both sides works equal except the one border case when the size of a file is 0.
//      In that case Progress works correctly without errors but Java implementation 
//      raises EndConditionException.
      
      if (data == null || data.isUnknown() || fname == null)
      {
         return;
      }

      Stream file = StreamFactory.openFileStream(fname, false, false);

      try
      {
         file.readBlock((memptr) data);
      }
      catch (EndConditionException | ErrorConditionException iioe)
      {
         throw new StopConditionException(iioe);
      }
      finally
      {
         file.close();
      }
   }

   /**
    * Read the full contents of the file and write that to the given binary object.
    *
    * @param    fname
    *           The target filename.
    * @param    data
    *           The buffer in which to write.
    */
   public static void readFromFile(character fname, BinaryData data)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      readFromFile(fname.toStringMessage(), data);
   }

   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(BinaryData data, character fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(BinaryData data, character fname, boolean append)
   {
      if (fname == null || fname.isUnknown())
      {
         return;
      }
      
      writeToFile(data, fname.toStringMessage(), append);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    */
   public static void writeToFile(BinaryData data, String fname)
   {
      writeToFile(data, fname, false);
   }
   
   /**
    * Write the full contents of the given binary object to the file specified by the given
    * filename.
    *
    * @param    data
    *           The data to write.
    * @param    fname
    *           The target filename.
    * @param    append
    *           <code>true</code> if the contents should be appended to the file.
    */
   public static void writeToFile(BinaryData data, String fname, boolean append)
   {
      // TODO: this is just a first pass, the real/full behavior needs to be determined
      
      if (data == null || data.isUnknown() || fname == null || fname.length() == 0)
      {
         return;
      }
      
      Stream file     = StreamFactory.openFileStream(fname, true, append);
      byte[] contents = data.getByteArray();
      
      try
      {
         // empty array should yield an empty file
         if (contents.length > 0)
         {
            file.write(contents);
         }
      }
      
      catch (InterruptedIOException iioe)
      {
         throw new StopConditionException(iioe);
      }
      
      catch (IOException ioe)
      {
         ErrorManager.showErrorAndAbend(140, Stream.PIPE_BROKEN_MESSAGE);
      }
      
      finally
      {
         file.close();
      }
   }
   
   /**
    * Raise a 12012 error when a <code>null</code> character is encountered.
    *
    * @param    data
    *           The bytes to check.
    *
    * @return   <code>true</code> if the error was raised. This may happen in a NO-ERROR
    *           case, which means the caller should not execute further.
    */
   private static boolean raiseErrorOnNullChar(byte[] data)
   {
      // TODO: the codepage name is hardcoded but should be resolved dynamically
      for (int i = 0; i < data.length; i++)
      {
         if (data[i] == 0)
         {
            String msg = "Invalid character data found in MEMPTR for codepage ISO8859-15";
            ErrorManager.recordOrThrowError(12012, msg);
            return true;
         }
      }
      
      return false;
   }
}
