/*
** Module   : LegacyLogManager.java
** Abstract : Define all APIs available via the LOG-MANAGER and DSLOG-MANAGER system handles.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20141028 Created initial version.
*/

package com.goldencode.p2j.util;

/**
 * Exposes the following methods or attributes from the LOG-MANAGER and DSLOG-MANAGER system
 * handles:
 * <table>
 * <tr>
 *    <th>Attribute or Method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>ENTRY-TYPES-LIST</td>
 *    <td>{@link #getEntryTypesList()}</td>
 * </tr>
 * <tr>
 *    <td>LOG-ENTRY-TYPES</td>
 *    <td>{@link #getLogEntryTypes()}, {@link #setLogEntryTypes(character)}</td>
 * </tr>
 * <tr>
 *    <td>LOGFILE-NAME</td>
 *    <td>{@link #getLogFileName()}, {@link #setLogFileName(character)}</td>
 * </tr>
 * <tr>
 *    <td>LOGGING-LEVEL</td>
 *    <td>{@link #getLoggingLevel()}, {@link #setLoggingLevel(integer)}</td>
 * </tr>
 * <tr>
 *    <td>LOG-THRESHOLD</td>
 *    <td>{@link #getLogThreshold()}</td>
 * </tr>
 * <tr>
 *    <td>NUM-LOG-FILES</td>
 *    <td>{@link #getNumLogFiles()}</td>
 * </tr>
 * <tr>
 *    <td>DB-CONTEXT</td>
 *    <td>{@link #getDBContext()}, {@link #setDBContext(character)}</td>
 * </tr>
 * <tr>
 *    <td>CLEAR-LOG()</td>
 *    <td>{@link #clearLog()}</td>
 * </tr>
 * <tr>
 *    <td>CLOSE-LOG()</td>
 *    <td>{@link #closeLog()}</td>
 * </tr>
 * <tr>
 *    <td>WRITE-MESSAGE()</td>
 *    <td>{@link #writeMessage(character)}, {@link #writeMessage(character, character)}</td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the {@link handle#unwrapLogManager}
 * API, which, depending on the handle's state, it either returns an 
 * {@link LegacyLogManager} instance or a dynamic proxy using
 * {@link handle#invalidAttrAccessProxy}. 
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public interface LegacyLogManager
extends CommonHandle,
        InstantiatingProcedure
{
   /**
    * Provides the comma-separated list of currently valid entry types that can be assigned to
    * the {@link #setLogEntryTypes}.
    *
    * @return   The list of valid log entry types.
    */
   @LegacyAttribute(name = "ENTRY-TYPES-LIST")
   public character getEntryTypesList();
   
   /**
    * Provides the comma-separated list of entry types that are being actively used for logging.
    *
    * @return   The list of active log entry types.
    */
   @LegacyAttribute(name = "LOG-ENTRY-TYPES")
   public character getLogEntryTypes();
   
   /**
    * Modifies the list of log entry types that are being actively used for logging.
    *
    * @param    types
    *           The comma-separated list of active entry types.
    */
   @LegacyAttribute(name = "LOG-ENTRY-TYPES", setter = true)
   public void setLogEntryTypes(String types);
   
   /**
    * Modifies the list of log entry types that are being actively used for logging.
    *
    * @param    types
    *           The comma-separated list of active entry types.
    */
   @LegacyAttribute(name = "LOG-ENTRY-TYPES", setter = true)
   public void setLogEntryTypes(character types);
   
   /**
    * Obtains the filename to which log entries and stack traces are written.
    *
    * @return   The filename.
    */
   @LegacyAttribute(name = "LOGFILE-NAME")
   public character getLogFileName();
   
   /**
    * Assigns the filename to which log entries and stack traces are written.
    *
    * @param    name
    *           The log file name.
    */
   @LegacyAttribute(name = "LOGFILE-NAME", setter = true)
   public void setLogFileName(String name);
   
   /**
    * Assigns the filename to which log entries and stack traces are written.
    *
    * @param    name
    *           The log file name.
    */
   @LegacyAttribute(name = "LOGFILE-NAME", setter = true)
   public void setLogFileName(character name);
   
   /**
    * Obtains the default logging level for all logging categories.
    *
    * @return   The logging level.
    */
   @LegacyAttribute(name = "LOGGING-LEVEL")
   public integer getLoggingLevel();
   
   /**
    * Assigns the default logging level for all logging categories.
    * <p>
    * There are 5 valid levels:
    * <p>
    * <ul>
    *   <li> 0 - no logging
    *   <li> 1 - only errors
    *   <li> 2 - basic, each entry type determines its own output but all types will generate
    *            some output
    *   <li> 3 - verbose
    *   <li> 4 - extended
    * </ul>
    *
    * @param    level
    *           The logging level to use by default.
    */
   @LegacyAttribute(name = "LOGGING-LEVEL", setter = true)
   public void setLoggingLevel(int level);
   
   /**
    * Assigns the default logging level for all logging categories.
    * <p>
    * There are 5 valid levels:
    * <p>
    * <ul>
    *   <li> 0 - no logging
    *   <li> 1 - only errors
    *   <li> 2 - basic, each entry type determines its own output but all types will generate
    *            some output
    *   <li> 3 - verbose
    *   <li> 4 - extended
    * </ul>
    *
    * @param    level
    *           The logging level to use by default.
    */
   @LegacyAttribute(name = "LOGGING-LEVEL", setter = true)
   public void setLoggingLevel(integer level);
   
   /**
    * Obtains the configured log file size in bytes at which the log file will be renamed and a
    * new log file started.
    *
    * @return   The maximum log file size.
    */
   @LegacyAttribute(name = "LOG-THRESHOLD")
   public integer getLogThreshold();
   
   /**
    * Obtains the configured number of log files that will be stored in total (including both
    * backed up log files and the current log).
    *
    * @return   The maximum number of log files.
    */
   @LegacyAttribute(name = "NUM-LOG-FILES")
   public integer getNumLogFiles();
   
   /**
    * Obtains the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @return   The logical database name or <code>unknown</code> if no database name has been
    *           assigned.
    */
   @LegacyAttribute(name = "DB-CONTEXT")
   public character getDBContext();
   
   /**
    * Assigns the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @param    ldbname
    *           The logical database name.
    */
   @LegacyAttribute(name = "DB-CONTEXT", setter = true)
   public void setDBContext(String ldbname);
   
   /**
    * Assigns the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @param    ldbname
    *           The logical database name.
    */
   @LegacyAttribute(name = "DB-CONTEXT", setter = true)
   public void setDBContext(character ldbname);
   
   /**
    * Delete all messages from the currently open log file.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "CLEAR-LOG")
   public logical clearLog();
   
   /**
    * Close the currently open log file.  Additional logging will not be possible after this
    * operation.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "CLOSE-LOG")
   public logical closeLog();
   
   /**
    * Write a user defined message to the currently open log file using the default subsystem
    * name "APPL".
    *
    * @param    msg
    *           The message to write.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(String msg);
   
   /**
    * Write a user defined message to the currently open log file using the default subsystem
    * name "APPL".
    *
    * @param    msg
    *           The message to write.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(character msg);
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(String msg, String subsys);
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(character msg, String subsys);
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(String msg, character subsys);
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   @LegacyMethod(name = "WRITE-MESSAGE")
   public logical writeMessage(character msg, character subsys);
}
