/*
** Module   : LegacyLogOps.java
** Abstract : Factory to create and manage the LOG-MANAGER and DSLOG-MANAGER instances.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20141029 Created initial version.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.security.*;

/**
 * Factory to create and manage the LOG-MANAGER and DSLOG-MANAGER instances. These instances
 * are created per-context so that each session has its own set, since in the 4GL the logging
 * will be different per client.
 */
public class LegacyLogOps
{
   /** Stores this context's log manager instances. */
   private static final ContextLocal<WorkArea> local =
   new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /**
    * Disables construction of instances of this object.  All access is via static methods.
    */
   private LegacyLogOps()
   {
   }
   
   /**
    * Returns the instance of the current session's <code>LOG-MANAGER</code> resource.
    * 
    * @return   The <code>LOG-MANAGER</code> instance.
    */
   public static LegacyLogManager logMgr()
   {
      WorkArea wa = local.get();
      
      LegacyLogManager lm = null;
      
      synchronized (wa)
      {
         if (wa.logMgr == null)
         {
            wa.logMgr = new LegacyLogManagerImpl(false);
         }
         
         lm = wa.logMgr;
      }
      
      return lm;
   }
   
   /**
    * Returns the instance of the current session's <code>LOG-MANAGER</code> system handle.
    * 
    * @return   The <code>LOG-MANAGER</code> system handle.
    */
   public static handle logMgrHandle()
   {
      return new handle(logMgr());
   }
   
   /**
    * Returns the instance of the current session's <code>DSLOG-MANAGER</code> resource.
    * 
    * @return   The <code>DSLOG-MANAGER</code> instance.
    */
   public static LegacyLogManager dsLogMgr()
   {
      WorkArea wa = local.get();
      
      LegacyLogManager lm = null;
      
      synchronized (wa)
      {
         if (wa.dsLogMgr == null)
         {
            wa.dsLogMgr = new LegacyLogManagerImpl(true);
         }
         
         lm = wa.dsLogMgr;
      }
      
      return lm;
   }
   
   /**
    * Returns the instance of the current session's <code>DSLOG-MANAGER</code> system handle.
    * 
    * @return   The <code>DSLOG-MANAGER</code> system handle.
    */
   public static handle dsLogMgrHandle()
   {
      return new handle(dsLogMgr());
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** Equivalent of the LOG-MANAGER system handle for this session. */
      private LegacyLogManager logMgr = null;
      
      /** Equivalent of the DSLOG-MANAGER system handle for this session. */
      private LegacyLogManager dsLogMgr = null;
   }
}
