/*
** Module   : TextOps.java
** Abstract : Progress 4GL compatible TextOps object
**
** Copyright (c) 2012-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 AIL 20121206 Created initial version by moving static methods from the character class as 
**                  this methods will be used with character and also longchar arguments.
** 002 CA  20130226 Added LENGTH function support when the type is an expression.
** 003 CA  20130306 Added missing numEntries(String, String).
** 004 OM  20130322 Added missing length forms.
** 005 VIG 20131010 Added null processing for type argument in length-methods.
** 006 OM  20140924 Added extra parameter to matches methods so they can handle unix-escapes as
**                  configured/converted. Fixed case-sensitivity of matches methods.
** 007 GES 20141003 Embedded null character handling fixes.
*/

package com.goldencode.p2j.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A class that provides Progress 4GL compatible character and longchar 
 * functions and operators.  This class provides all functionality as statics
 * and has no instance data.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * +                operator (concatenation) {@link #concat}
 * num-entries      function                 {@link #numEntries}
 * begins           operator                 {@link #begins} and {@link #_begins}
 * matches          operator                 {@link #matches} and {@link #_matches} 
 * can-do           function                 {@link #matchesList}
 * caps             function                 {@link #toUpperCase}
 * compare          function                 {@link #compare}
 * entry            function                 {@link #entry}
 * index            function                 {@link #indexOf}
 * lc               function                 {@link #toLowerCase}
 * left-trim        function                 {@link #leftTrim}
 * length           function                 {@link #length} and {@link #byteLength}
 * lookup           function                 {@link #lookup}
 * maximum          function                 {@link #maximum}
 * minimum          function                 {@link #minimum}
 * r-index          function                 {@link #lastIndexOf}
 * replace          function                 {@link #replaceAll}
 * right-trim       function                 {@link #rightTrim}
 * substitute       function                 {@link #substitute}
 * substring        function                 {@link #substring}
 * trim             function                 {@link #trim}
 * entry            language statement       {@link Text#replaceEntry}
 * overlay          language statement       {@link Text#overlay}
 * substring        language statement       {@link Text#replace}
 * </pre>
 * <p>
 * There are convenience versions of the begins and matches operators that
 * directly return <code>boolean</code> values such that unwrapping is not
 * needed.  These are named the same as the operator methods above except 
 * there is no 'With' appended.  For example, the begins operator is
 * {@link #begins}.
 * <p>
 * The following is the list of Progress language features which optionally
 * return a <code>character</code> instance that reflects the case-sensitivity
 * of the operands or input values:
 * <pre>
 * +                operator
 * caps             function                       {@link #toUpperCase}
 * lc               function                       {@link #toLowerCase}
 * left-trim        function                       {@link #leftTrim} 
 * maximum          function                       {@link #maximum}
 * minimum          function                       {@link #minimum}
 * replace          function                       {@link #replaceAll}
 * right-trim       function                       {@link #rightTrim} 
 * substring        function                       {@link #substring}
 * trim             function                       {@link #trim}
 * </pre>
 * <p>
 * Strangely enough, the <code>chr, entry</code> and <code>substitute</code>
 * functions DO NOT provide this support in Progress.  Each of the
 * corresponding backing methods properly duplicate this behavior.
 * <p>
 * This class implements unknown value logic for all operators and functions.
 * Generally, if any parameter of the these methods is unknown, the result is
 * unknown.
 */
public final class TextOps
{
   /** Specifies the less than operator for {@link #compare}. */
   private static final int LT = 0;
   
   /** Specifies the less than or equal operator for {@link #compare}. */
   private static final int LE = 1;
   
   /** Specifies the greater than operator for {@link #compare}. */
   private static final int GT = 2;
   
   /** Specifies the greater than or equal operator {@link #compare}. */
   private static final int GE = 3;
   
   /** Specifies the equal operator for {@link #compare}. */
   private static final int EQ = 4;
   
   /** Specifies the not equal operator for {@link #compare}. */
   private static final int NE = 5;

   /** Specifies the matches operator for {@link #compare}. */
   private static final int MATCHES = 6;

   /** Specifies the begins operator for {@link #compare}. */
   private static final int BEGINS = 7;
   
   /** Specifies the RAW type for {@link #compare} or {@link #substring}. */
   private static final int RAW = 6;
   
   /** 
    * Specifies the COLUMN type for {@link #compare} or {@link #substring}.
    */
   private static final int COLUMN = 7;
   
   /** Specifies the FIXED type for {@link #compare} or {@link #substring}. */
   private static final int FIXED = 8;
   
   /**
    * Specifies the CHARACTER type for {@link #compare} or {@link #substring}.
    */      
   private static final int CHARACTER = 9;
   
   /** Specifies the CASE-SENSITIVE strength for {@link #compare}. */
   private static final int CASE_SENSITIVE = 10;
   
   /** Specifies the CASE-INSENSITIVE strength for {@link #compare}. */
   private static final int CASE_INSENSITIVE = 11;
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *          
    * @return  <code>true</code> if op matches the regular expression or <code>false</code> if
    *          there is no match or if either operand is <code>unknown</code>.
    */
   public static boolean _matches(Text op, character pattern, boolean windows)
   {
      if (op.isUnknown() || pattern.isUnknown())
      {
         return false;
      }
      
      boolean caseSensitive = op.isCaseSensitive() || pattern.isCaseSensitive();
      return _matches(op.getValue(), pattern.getValue(), caseSensitive, windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * basing case sensitivity on the left operand.
    * <p>
    * This form of the method is unix-escape independent because Progress style matching pattern
    * passed as a parameter is already transformed at conversion time to a regular expression.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          RegExp matching pattern transformed from Progress at conversion time.
    *
    * @return  <code>true</code> if op matches the regular expression or <code>false</code> if 
    *          there is no match or if the left operand is <code>unknown</code>.
    */
   public static boolean _matches(Text op, String pattern)
   {
      if (op.isUnknown())
      {
         return false;
      }
      
      String string = op.getValue();
      if (!op.isCaseSensitive())
      {
         string = string.toUpperCase();
         pattern = pattern.toUpperCase();
      }
      
      return string.matches(pattern);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * case-insensitively.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular 
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression or <code>false</code> if
    *          there is no match or if the right operand is <code>unknown</code>.
    */
   public static boolean _matches(String op, character pattern, boolean windows)
   {
      if (pattern.isUnknown())
      {
         return false;
      }
      
      return _matches(op, pattern.getValue(), pattern.isCaseSensitive(), windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * ignoring the case of the operand and the pattern.
    * <p>
    * This form of the method is unix-escape independent because Progress style matching pattern
    * passed as a parameter is already transformed at conversion time to a regular expression.
    * By default, in P4GL character operations are case insensitive.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          RegExp matching pattern transformed from Progress at conversion time.
    * 
    * @return   <code>true</code> if op matches the regular expression.
    */
   public static boolean _matches(String op, String pattern)
   {
      op = op.toUpperCase();
      pattern = pattern.toUpperCase();
      
      return op.matches(pattern);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller, 
    * honoring the case sensitivity preference of the caller.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   caseSens
    *          <code>true</code> if a case-sensitive comparison should be used.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression.
    */
   public static boolean _matches(String op, String pattern, boolean caseSens, boolean windows)
   {
      String up = caseSens ? op : op.toUpperCase();
      
      pattern = caseSens ? pattern : pattern.toUpperCase();
      return up.matches(SQLHelper.convertToRegEx(pattern, windows));
   }
   
   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if either of the two operands is case-sensitive.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2 or <code>false</code> if there is no
    *          match or if either operand is <code>unknown value</code>.
    */
   public static boolean _begins(Text op1, character op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return false;
      }
      
      boolean cs = op1.isCaseSensitive() || op2.isCaseSensitive();
      
      return _begins(op1.getValue(), op2.getValue(), cs);
   }
   
   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if the right operand is case-sensitive.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2 or <code>false</code> if there is no 
    *          match or if the right operand is <code>unknown value</code>.
    */
   public static boolean _begins(String op1, character op2)
   {
      if (op2.isUnknown())
      {
         return false;
      }
      
      return _begins(op1, op2.getValue(), op2.isCaseSensitive());
   }
   
   /**
    * Tests if the first operand begins with the second operand, implementing a case-sensitive
    * comparison if the left operand is case-sensitive.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2 or <code>false</code> if there is no 
    *          match or if the left operand is <code>unknown value</code>.
    */
   public static boolean _begins(Text op1, String op2)
   {
      if (op1.isUnknown())
      {
         return false;
      }
      
      return _begins(op1.getValue(), op2, op1.isCaseSensitive());
   }
   
   /**
    * Tests if the first operand begins with the second operand, ignoring the case of the two
    * operands.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2.
    */
   public static boolean _begins(String op1, String op2)
   {
      return _begins(op1, op2, false);
   }
   
   /**
    * Tests if the first operand begins with the second operand, honoring the case sensitivity
    * preference of the caller.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    * @param   caseSens
    *          <code>true</code> if a case-sensitive comparison should be used.
    *
    * @return  <code>true</code> if op1 begins with op2.
    */
   public static boolean _begins(String op1, String op2, boolean caseSens)
   {
      String up1 = caseSens ? op1 : op1.toUpperCase();
      String up2 = caseSens ? op2 : op2.toUpperCase();
      
      return up1.startsWith(up2);
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if either of the two operands is
    * case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>unknown value</code> if either operand is
    *           <code>unknown value</code>.
    */
   public static logical begins(Text op1, character op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new logical();
      }
      
      boolean cs = (op1.isCaseSensitive() || op2.isCaseSensitive());
      
      return begins(op1.getValue(), op2.getValue(), cs);
   }
   
   /**
    * Tests if the first operand begins with the second operand, implementing a case-sensitive
    * comparison if the right operand is case-sensitive.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2 or <code>unknown value</code> if the
    *          right operand is <code>unknown value</code>.
    */
   public static logical begins(String op1, character op2)
   {
      if (op2.isUnknown())
      {
         return new logical();
      }
      
      return begins(op1, op2.getValue(), op2.isCaseSensitive());
   }
   
   /**
    * Tests if the first operand begins with the second operand, implementing a case-sensitive
    * comparison if the left operand is case-sensitive.
    *
    * @param   op1
    *          First string to compare.
    * @param   op2
    *          Second string to compare.
    *
    * @return  <code>true</code> if op1 begins with op2 or <code>unknown value</code> if the left
    *          operand is <code>unknown value</code>.
    */
   public static logical begins(Text op1, String op2)
   {
      if (op1.isUnknown())
      {
         return new logical();
      }
      
      return begins(op1.getValue(), op2, op1.isCaseSensitive());
   }
   
   /**
    * Tests if the first operand begins with the second operand, ignoring the
    * case of the two operands.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static logical begins(String op1, String op2)
   {
      return begins(op1, op2, false);
   }

   /**
    * Tests if the first operand begins with the second operand, honoring the
    * case sensitivity preference of the caller.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    * @param    caseSens
    *           <code>true</code> if a case-sensitive comparison should be used.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static logical begins(String op1, String op2, boolean caseSens)
   {
      String up1 = caseSens ? op1 : op1.toUpperCase();
      String up2 = caseSens ? op2 : op2.toUpperCase();
      
      return new logical(up1.startsWith(up2));
   }

   /**
    * Returns the length of the given <code>character</code> or 
    * <code>longchar</code> instance in bytes.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of bytes in the given instance or -1 if the
    *           instance is <code>unknown value</code>.
    */
   public static int byteLengthOf(Text c)
   {
      if (c.isUnknown())
         return -1;
      
      return c.getValue().getBytes().length;
   }

   /**
    * Returns the length of the given <code>character</code> literal in bytes.
    *
    * @param    s
    *           The literal to test.
    *
    * @return   The number of bytes in the given literal.
    */
   public static integer byteLength(String s)
   {
      return new integer(s.getBytes().length);
   }

   /**
    * Returns the length of the given <code>character</code> or 
    * <code>longchar</code> instance in bytes.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of bytes in the given instance or
    *           <code>unknown value</code> if the instance is
    *           <code>unknown value</code>.
    */
   public static integer byteLength(Text c)
   {
      if (c.isUnknown())
         return new integer();
      
      return new integer(c.getValue().getBytes().length);
   }
   
   /**
    * Convert the given text from one code page to another. As no codepage is
    * specified for this API, no conversion occurs.
    * 
    * @param    text
    *           The text which needs to be converted.
    *           
    * @return   The same text, with no change.
    */
   public static <T extends Text> T codePageConvert(T text)
   {
      return text;
   }
   
   /**
    * Convert the given text using the specified codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T text, String target)
   {
      return codePageConvert(text, target, (String) null);
   }
   
   /**
    * Convert the given text using the specified codepage, assuming the text
    * is encoded with the given source codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    * @param    source
    *           The source codepage. If the text is a {@link longchar} 
    *           instance, then the text's internal codepage is used.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T      text,
                                                    String target,
                                                    String source)
   {
      // TODO: validate target and source
      // TODO: IMPLEMENT THIS
      return text;
   }
   
   /**
    * Convert the given text using the specified codepage, assuming the text
    * is encoded with the given source codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    * @param    source
    *           The source codepage. If the text is a {@link longchar} 
    *           instance, then the text's internal codepage is used.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T         text, 
                                                    String    target,
                                                    character source)
   {
      // TODO: validate source
      return codePageConvert(text, target, source.getValue());
   }
   
   /**
    * Convert the given text using the specified codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T text, character target)
   {
      // TODO: validate target
      return codePageConvert(text, target.getValue(), (String) null);
   }
   
   /**
    * Convert the given text using the specified codepage, assuming the text
    * is encoded with the given source codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    * @param    source
    *           The source codepage. If the text is a {@link longchar} 
    *           instance, then the text's internal codepage is used.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T         text,
                                                    character target,
                                                    String    source)
   {
      // TODO: validate target
      return codePageConvert(text, target.getValue(), source);
   }
   
   /**
    * Convert the given text using the specified codepage, assuming the text
    * is encoded with the given source codepage.
    * 
    * @param    text
    *           The text to be converted.
    * @param    target
    *           The target codepage.
    * @param    source
    *           The source codepage. If the text is a {@link longchar} 
    *           instance, then the text's internal codepage is used.
    *           
    * @return   The converted text.
    */
   public static <T extends Text> T codePageConvert(T         text,
                                                    character target,
                                                    character source)
   {
      // TODO: validate target and source
      return codePageConvert(text, target.getValue(), source.getValue());
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 character strength)
   {
      return compare(op1, operator, op2, strength, (character) null);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 String    operator,
                                 Text      op2,
                                 character strength)
   {
      return compare(op1, new character(operator), op2, strength);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 String    strength)
   {
      return compare(op1, operator, op2, new character(strength));
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text   op1,
                                 String operator,
                                 Text   op2,
                                 String strength)
   {
      return compare(op1, 
                     new character(operator), 
                     op2,
                     new character(strength));
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    * 
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 character strength,
                                 character collation)
   {
      int op = getCompareOperator(operator);
      int str = getStrength(strength);
      
      if (op < 0 || str < 0)
      {
         // TODO: what is the correct returned value ?
         return new logical();
      }
      
      // TODO: validate collation, if not-null
      
      return compare(op1, op, op2, str, collation.getValue());
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 String    operator,
                                 Text      op2,
                                 character strength,
                                 character collation)
   {
      return compare(op1, new character(operator), op2, strength, collation);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 String    strength,
                                 character collation)
   {
      return compare(op1, operator, op2, new character(strength), collation);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 String    operator,
                                 Text      op2,
                                 String    strength,
                                 character collation)
   {
      return compare(op1, 
                     new character(operator), 
                     op2,
                     new character(strength),
                     collation);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    * 
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 character strength,
                                 String    collation)
   {
      character chcol = collation == null ? (character) null 
                                          : new character(collation);

      return compare(op1, operator, op2, strength, chcol);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 String    operator,
                                 Text      op2,
                                 character strength,
                                 String    collation)
   {
      return compare(op1, new character(operator), op2, strength, collation);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text      op1,
                                 character operator,
                                 Text      op2,
                                 String    strength,
                                 String    collation)
   {
      return compare(op1, operator, op2, new character(strength), collation);
   }

   /**
    * Compare the two texts using the given operator and the given text.
    * The operator must evaluate to one of the following relational operators:
    * <ul>
    * <li> LT (or &lt;)
    * <li> LE (or &lt;=)
    * <li> EQ (or =)
    * <li> GE (or &gt;=)
    * <li> GT (or &gt;)
    * <li> NE (or &lt;&gt;)
    * <li> BEGINS
    * <li> MATCHES
    * </ul>
    * The strength must evaluate to one of the given values:
    * <ul>
    * <li> RAW - compares the two strings using the numeric values in the 
    *      current code page.
    * <li> CASE-SENSITIVE - performs a case-sensitive comparison 
    * <li> CASE-INSENSITIVE - performs a case-insensitive comparison
    * <li> CAPS - converts any lowercase letters in the two strings to 
    *      uppercase letters, and then performs a raw comparison of the 
    *      resulting strings. 
    *      When neither string contains a wildcard character, this option 
    *      behaves the same as the MATCHES operator.
    *       
    * @param    op1
    *           The first string to be compared.
    * @param    operator
    *           The operator to be used for this comparison. Must evaluate to
    *           one of the strings presented above.
    * @param    op2
    *           The second string to be compared.
    * @param    strength
    *           The used strength. Must evaluate to one of the strings 
    *           presented above.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *            
    * @return   A {@link logical} value representing the result of the 
    *           comparison, depending on the operator and strength used.
    */
   public static logical compare(Text   op1,
                                 String operator,
                                 Text   op2,
                                 String strength,
                                 String collation)
   {
      return compare(op1, 
                     new character(operator), 
                     op2,
                     new character(strength),
                     collation);
   }

   /**
    * Compares two character values with a runtime-specified strength and
    * operator.  This method will probably generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for conversion.
    * The result should match up with the Progress equivalents in the ASCII
    * range.
    * <p>
    * This method does not handle the case of user-specified collation tables
    * as is possible in Progress.
    * <p>
    * In addition, the case-sensitive and case-insensitive strength options
    * are likely to differ in some results since the current implementation
    * really just implements a direct numeric comparison based on each 
    * UNICODE character's value (though case-insensitive support is
    * provided but it just converts everything to the same case before
    * implementing the 'raw' lexicographic comparison.
    *
    * @param    op1
    *           The left operand.
    * @param    operator
    *           One of {@link TextOps#LT}, {@link TextOps#LE},
    *           {@link TextOps#GT}, {@link TextOps#GE},
    *           {@link TextOps#EQ} or {@link TextOps#NE}.
    * @param    op2
    *           The right operand.
    * @param    strength
    *           One of {@link TextOps#RAW}, {@link TextOps#CASE_SENSITIVE}
    *           or {@link TextOps#CASE_INSENSITIVE}.
    *           
    * @return   The result of the condition evaluation. 
    */   
   public static logical compare(Text   op1,
                                 int    operator,
                                 Text   op2,
                                 int    strength)
   {
      return compare(op1, operator, op2, strength, null);
   }
   
   /**
    * Compares two character values with a runtime-specified strength and
    * operator.  This method will probably generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for conversion.
    * The result should match up with the Progress equivalents in the ASCII
    * range.
    * <p>
    * This method does not handle the case of user-specified collation tables
    * as is possible in Progress.
    * <p>
    * In addition, the case-sensitive and case-insensitive strength options
    * are likely to differ in some results since the current implementation
    * really just implements a direct numeric comparison based on each 
    * UNICODE character's value (though case-insensitive support is
    * provided but it just converts everything to the same case before
    * implementing the 'raw' lexicographic comparison.
    *
    * @param    op1
    *           The left operand.
    * @param    operator
    *           One of {@link TextOps#LT}, {@link TextOps#LE},
    *           {@link TextOps#GT}, {@link TextOps#GE},
    *           {@link TextOps#EQ} or {@link TextOps#NE}.
    * @param    op2
    *           The right operand.
    * @param    strength
    *           One of {@link TextOps#RAW}, {@link TextOps#CASE_SENSITIVE}
    *           or {@link TextOps#CASE_INSENSITIVE}.
    * @param    collation
    *           The used collation. When <code>null</code>, the default 
    *           collation is used.
    *           
    * @return   The result of the condition evaluation. 
    */   
   public static logical compare(Text   op1,
                                 int    operator,
                                 Text   op2,
                                 int    strength,
                                 String collation)
   {
      // TODO: add collation support
      // TODO: add support for MATCHES and BEGINS operators

      if (operator != LT && operator != LE && operator != GT &&
          operator != GE && operator != EQ && operator != NE &&
          operator != MATCHES && operator != BEGINS) 
      {
         // unrecognized operator
         return new logical();
      }
      
      // unknown value logic
      if (op1.isUnknown() && op2.isUnknown())
      {
         // both are unknown
         if (operator == LE || operator == GE || operator == EQ) 
         {
            return new logical(true);
         }
         else
         {
            return new logical(false);
         }
      }
      else
      {
         // both are NOT unknown BUT...
         if (op1.isUnknown() || op2.isUnknown())
         {
            // one or the other is unknown
            if (operator == NE)
            {
               return new logical(true);
            }
            else
            {
               return new logical(false);
            }
         }
      }
      
      int result;
      
      // TODO: add support for the other strength typeS: CAPS, PRIMARY, 
      // SECONDARY, TERTIARY, QUATERNARY

      if (strength == RAW || strength == CASE_SENSITIVE)
      {
         // RAW and CASE_SENSITIVE are the same right now
         result = op1.getValue().compareTo(op2.getValue());
      }
      else if (strength == CASE_INSENSITIVE) 
      {
         result = op1.getValue().compareToIgnoreCase(op2.getValue());
      }
      else
      {
         // unrecognized strength
         return new logical();
      }
      
      boolean conclude = false;
      
      switch (operator)
      {
         case LT:
            conclude = (result < 0);
            break;
         case LE:
            conclude = (result <= 0);
            break;
         case GT:
            conclude = (result > 0);
            break;
         case GE:
            conclude = (result >= 0);
            break;
         case EQ:
            conclude = (result == 0);
            break;
         case NE:
            conclude = (result != 0);
            break;
      }
      
      return new logical(conclude);
   }

   /**
    * Implements the <code>+</code> character concatination operator which 
    * returns an instance of the same class.  If any operand is the
    * <code>unknown value</code>, the result is <code>unknown value</code>.
    * If any operand is case-sensitive, the result will be case-sensitive.
    * <p>
    * It is valid to pass a mixture of <code>BaseDataType</code> and other
    * object types to this method.  Types that are not <code>character</code>
    * will not affect the case-sensitivity or unknown status of the result.
    * 
    * TODO: this should return Text, but for now we let it return character, as this can be used
    * with lots of builtin functions/methods which expect a character parameter, not a Text 
    * parameter.
    * 
    * @param    list
    *           The operands.
    *
    * @return   The result of the operator.
    */
   public static character concat(Object ... list)
   {
      // TODO: add longchar support
      StringBuilder sb = new StringBuilder();
      
      boolean caseSens = false;
      boolean first    = true;
      
      for (Object txt : list)
      {
         if (txt instanceof character)
         {
            character op = (character) txt;
            
            // any character operand that is unknown causes the entire
            // result to be unknown
            if (op.isUnknown())
            {
               return new character();
            }
            
            // any character operand that is case-sensitive causes the
            // entire result to be case-sensitive
            caseSens = (caseSens || op.isCaseSensitive());
            
            // the leftmost operand does not truncate at the embedded null, but all subsequent
            // operands do (the safeValue() call truncates at the embedded null char)
            sb.append(first ? op.getValue() : op.safeValue());
         }
         else
         {
            sb.append(txt.toString());
         }
         
         first = false;
      }
      
      return new character(sb.toString(), caseSens);
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(NumberType i,
                                          T          list,
                                          character  delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown() || delimit.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(),
                           i.intValue(),
                           list.getValue(),
                           safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(double i, T list, character delimit)
   throws ErrorConditionException
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(),
                           (int) i, 
                           list.getValue(),
                           safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list, character delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || delimit.isUnknown())
      {
         return new character();
      }
      
      return entryImpl(character.class, 
                       i.intValue(),
                       list,
                       safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(NumberType i,
                                          T          list,
                                          String delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(), i.intValue(),
                           list.getValue(),
                           safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(double i, T list, String delimit)
   throws ErrorConditionException
   {
      if (list.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(),
                           (int) i,
                           list.getValue(),
                           safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list, String delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         return new character();
      }
      
      return entryImpl(character.class,
                       i.intValue(),
                       list, 
                       safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(NumberType i, T list)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(),
                           i.intValue(),
                           list.getValue(),
                           ',');
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         return new character();
      }
      
      return entryImpl(character.class, i.intValue(), list, ',');
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static <T extends Text> T entry(double i, T list)
   throws ErrorConditionException
   {
      if (list.isUnknown())
      {
         return (T) instantiateUnknown(list.getClass());
      }
      
      return (T) entryImpl(list.getClass(), (int) i, list.getValue(), ',');
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, String list)
   throws ErrorConditionException
   {
      return entry((int) i, list, ',');
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, String list, String delimit)
   throws ErrorConditionException
   {
      return entryImpl(character.class,
                       (int) i, 
                       list,
                       safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, String list, char delimit)
   throws ErrorConditionException
   {
      return entryImpl(character.class, (int) i, list, delimit);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, Text target, NumberType i)
   {
      if (src.isUnknown() || target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     i.doubleValue(),
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String src, Text target, NumberType i)
   {
      if (target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src,
                     target.getValue(),
                     i.doubleValue(),
                     target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, String target, NumberType i)
   {
      if (src.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target,
                     i.doubleValue(),
                     src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String src, String target, NumberType i)
   {
      if (i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target, i.doubleValue(), false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, Text target, double i)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     i,
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String src, Text target, double i)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target.getValue(), i, target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, String target, double i)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(), target, i, src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, Text target)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     1,
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String src, Text target)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target.getValue(), 1, target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(Text src, String target)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(), target, 1, src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The (1-based) starting index from which to search.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src, String target, double ii)
   {
      return indexOf(src, target, ii, false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src, String target)
   {
      return indexOf(src, target, 1, false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The (1-based) starting index from which to search.
    * @param    caseSensitive
    *           <code>true</code> if case must be considered in the search;
    *           <code>false</code> if the search is case insensitive.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src,
                                 String target,
                                 double ii,
                                 boolean caseSensitive)
   {
      int i = (int) ii;
      
      if (src.length() == 0 || target.length() == 0 || i < 1)
      {
         return new integer(0);
      }
      
      if (!caseSensitive)
      {
         src = src.toLowerCase();
         target = target.toLowerCase();
      }
      
      return new integer(src.indexOf(target, i - 1) + 1);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, Text target, NumberType i)
   {
      if (src.isUnknown() || target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue(), i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String src, Text target, NumberType i)
   {
      if (target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue(), i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, String target, NumberType i)
   {
      if (src.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target, i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String src, String target, NumberType i)
   {
      if (i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target, i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, Text target, double i)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue(), (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String src, Text target, double i)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue(), (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, String target, double i)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target, (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, Text target)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String src, Text target)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(Text src, String target)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer lastIndexOf(String src, String target, double ii)
   {
      int i = (int) ii;
      
      if (src.length() == 0 || target.length() == 0 || i < 1)
      {
         return new integer(0);
      }
      
      return new integer(src.lastIndexOf(target, i - 1) + 1);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer lastIndexOf(String src, String target)
   {
      return lastIndexOf(src, target, src.length() + 1);
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T leftTrim(T input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) leftTrimImpl(input.getClass(),
                              input.getValue(),
                              list.getValue(),
                              input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * the input parameter. If the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T leftTrim(T input, String list)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) leftTrimImpl(input.getClass(),
                              input.getValue(),
                              list,
                              input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed case-insensitively.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return leftTrimImpl(character.class, input, list.getValue(), false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T leftTrim(T input)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) leftTrimImpl(input.getClass(),
                              input.getValue(),
                              null,
                              input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String  input,
                                    String  list,
                                    boolean caseSens)
   {
      return leftTrimImpl(character.class,
                          input,
                          list,
                          caseSens);
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input, String list)
   {
      return leftTrim(input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left of an input string.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input)
   {
      return leftTrim(input, (String) null);
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance or -1 if the
    *           instance is <code>unknown value</code>.
    */
   public static int lengthOf(Text c)
   {
      if (c.isUnknown())
         return -1;
      
      return c.getValue().length();
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance
    *           <code>unknown value</code> if the instance is
    *           <code>unknown value</code>.
    */
   public static integer length(Text c)
   {
      if (c.isUnknown())
         return new integer();
      
      return new integer(c.getValue().length());
   }

   /**
    * Returns the length of the given <code>character</code> instance, using the type parameter
    * to determine the measurement.
    *
    * @param    c
    *           The instance to test.
    * @param    type
    *           The measurement mode.
    *           "character" is supposed to be the default measurement unit 
    *           for the LENGTH built-in function (when type is null)
    *
    * @return   The length of the given text
    */
   public static integer length(Text c, character type)
   {
      if (type == null)
      {
         return length(c, DynamicOps.CHAR_TYPE);
      }
      
      if (type.isUnknown())
      {
         return new integer();
      }

      return length(c, type.toStringMessage());
   }
   
   /**
    * Returns the length of the given <code>character</code> instance, using the type parameter
    * to determine the measurement.
    *
    * @param    c
    *           The instance to test.
    * @param    type
    *           The measurement mode.
    *           "character" is supposed to be the default measurement unit 
    *           for the LENGTH built-in function (when type is null)
    *
    * @return   The length of the given text
    */
   public static integer length(Text c, String type)
   {
      if (c == null || c.isUnknown())
      {
         return new integer();
      }
      
      if (type == null)
      {
         type = DynamicOps.CHAR_TYPE;
      }
      
      return TextOps.length(c.getValue(), type);
   }

   /**
    * Returns the length of the given <code>character</code> instance, using the type parameter
    * to determine the measurement.
    *
    * @param    c
    *           The instance to test.
    * @param    type
    *           The measurement mode.
    *           "character" is supposed to be the default measurement unit 
    *           for the LENGTH built-in function (when type is null)
    *
    * @return   The length of the given text
    */
   public static integer length(String c, character type)
   {
      if (type == null)
      {
         return TextOps.length(c, DynamicOps.CHAR_TYPE);
      }
      
      if (type.isUnknown())
      {
         return new integer();
      }
      
      return TextOps.length(c, type.toStringMessage());
   }

   /**
    * Returns the length of the given string instance in characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance.
    */
   public static integer length(String c)
   {
      return new integer(c.length());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, Text list, character delim)
   {
      if (c.isUnknown() || list.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(),
                    list.getValue(),
                    safeDelimiter(delim.getValue()),
                    cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list, character delim)
   {
      if (c.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(),
                    list,
                    safeDelimiter(delim.getValue()),
                    c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, Text list, character delim)
   {
      if (list.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c,
                    list.getValue(), 
                    safeDelimiter(delim.getValue()),
                    list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, Text list, String delim)
   {
      if (c.isUnknown() || list.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(), list.getValue(), safeDelimiter(delim), cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, Text list)
   {
      if (c.isUnknown() || list.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(), list.getValue(), ',', cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list)
   {
      if (c.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(), list, ',', c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, Text list)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list.getValue(), ',', list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter.  The search will be done
    * case-insensitively.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found.
    */
   public static integer lookup(String c, String list, char delim)
   {
      return lookup(c, list, delim, false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter.  
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    * @param    cs
    *           <code>true</code> specifies the match is to be done with
    *           case-sensitivity.
    *
    * @return   The 1-based index into the list or 0 if no match is found.
    */
   public static integer lookup(String c, String list, char delim, boolean cs)
   {
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delim);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // only use up to the first instance of the delimiter in the search 
      // pattern
      int    idx    = c.indexOf(delim);
      String search = idx < 0 ? c : c.substring(0, idx); 
      
      // default return value
      int index  = 0;
      
      // do our search
      for (int i = 0; i < array.length; i++)
      {
         if (cs ? array[i].equals(search) : array[i].equalsIgnoreCase(search))
         {
            // match found!  convert from 0-based to 1-based
            index = i + 1;
            break;
         }
      }
      
      return new integer(index);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be done case-insensitively.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list)
   {
      return lookup(c, list, ',', false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list, String delim)
   {
      return lookup(c, list, safeDelimiter(delim), false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, Text list, String delim)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c,
                    list.getValue(),
                    safeDelimiter(delim),
                    list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list, String delim)
   {
      if (c.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(),
                    list,
                    safeDelimiter(delim),
                    c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list, character delim)
   {
      if (delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list, safeDelimiter(delim.getValue()), false);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression or <code>unknown</code>
    *          value if either operand is <code>unknown</code>.
    */
   public static logical matches(Text op, character pattern, boolean windows)
   {
      if (op.isUnknown() || pattern.isUnknown())
      {
         return new logical();
      }
      
      boolean caseSensitive = op.isCaseSensitive() || pattern.isCaseSensitive();
      return matches(op.getValue(), pattern.getValue(), caseSensitive, windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression or <code>unknown</code>
    *          value if the left operand is <code>unknown</code>.
    */
   public static logical matches(Text op, String pattern, boolean windows)
   {
      if (op.isUnknown())
      {
         return new logical();
      }
      
      return matches(op.getValue(), pattern, op.isCaseSensitive(), windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * case-insensitively.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression or <code>unknown</code>
    *          value if the right operand is <code>unknown</code>.
    */
   public static logical matches(String op, character pattern, boolean windows)
   {
      if (pattern.isUnknown())
      {
         return new logical();
      }
      
      return matches(op, pattern.getValue(), pattern.isCaseSensitive(), windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * ignoring the case of the operand and the pattern.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression.
    */
   public static logical matches(String op, String pattern, boolean windows)
   {
      return matches(op, pattern, false, windows);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller,
    * honoring the case sensitivity preference of the caller.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted to a regular
    * expression using {@link SQLHelper#convertToRegEx} during this method.
    *
    * @param   op
    *          String to check for a match.
    * @param   pattern
    *          Progress style matching pattern.
    * @param   caseSens
    *          <code>true</code> if a case-sensitive comparison should be used.
    * @param   windows
    *          <code>true</code> if the backslash does not have special meaning. On Unix, P4GL
    *          the backslash (\)is an escape character, like tilde (\).
    *
    * @return  <code>true</code> if op matches the regular expression.
    */
   public static logical matches(String op, String pattern, boolean caseSens, boolean windows)
   {
      String up = caseSens ? op : op.toUpperCase();
      pattern = caseSens ? pattern : pattern.toUpperCase();
      
      boolean result = up.matches(SQLHelper.convertToRegEx(pattern, windows));
      return new logical(result);
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller.
    * <p>
    * This form of the method is unix-escape independent because Progress style matching pattern
    * passed as a parameter is already transformed at conversion time to a regular expression.
    * By default, in P4GL character operations are case insensitive.
    *
    * @param   op
    *          String to check for a match.
    * @param   regExp
    *          The RegExp pattern as obtained using {@link SQLHelper#convertToRegEx} from a
    *          Progress style matching pattern known at conversion time.
    *
    * @return  <code>true</code> if op matches the regular expression.
    */
   public static logical matches(String op, String regExp)
   {
      op = op.toUpperCase();
      regExp = regExp.toUpperCase();
      
      return new logical(op.matches(regExp));
   }
   
   /**
    * Tests if the first operand matches a regular expression that is provided by the caller.
    * <p>
    * This form of the method is unix-escape independent because Progress style matching pattern
    * passed as a parameter is already transformed at conversion time to a regular expression.
    *
    * @param   op
    *          String to check for a match.
    * @param   regExp
    *          The RegExp pattern as obtained using {@link SQLHelper#convertToRegEx} from a
    *          Progress style matching pattern known at conversion time.
    *
    * @return  <code>true</code> if op matches the regular expression.
    */
   public static logical matches(Text op, String regExp)
   {
      if (op.isUnknown())
      {
         return new logical();
      }
      
      String toMatch = op.getValue(); 
      if (!op.isCaseSensitive())
      {
         toMatch = toMatch.toUpperCase();
         regExp = regExp.toUpperCase();
      }
      return new logical(toMatch.matches(regExp));
   }
   
   /**
    * Tests if the current PROGRESS user is allowed in the list.
    * Since in the P2J environment the security subsystem is based on the
    * Security Manager, this call simply catches the overlooked CAN-DO call
    * in the PROGRESS code.
    *
    * @param    list
    *           comma-separated list of user IDs, possibly with ! and *
    *           modifiers.
    *
    * @return   never
    * @throws   IllegalArgumentException
    *           to catch the untranslated CAN-DO call
    */
   public static logical matchesList(Text list)
   throws IllegalArgumentException
   {
      throw new IllegalArgumentException("untranslated CAN-DO call");
   }

   /**
    * Tests if the current PROGRESS user is allowed in the list.
    * Since in the P2J environment the security subsystem is based on the
    * Security Manager, this call simply catches the overlooked CAN-DO call
    * in the PROGRESS code.
    *
    * @param    list
    *           comma-separated list of user IDs, possibly with ! and *
    *           modifiers.
    *
    * @return   never
    * @throws   IllegalArgumentException
    *           to catch the untranslated CAN-DO call
    */
   public static logical matchesList(String list)
   throws IllegalArgumentException
   {
      throw new IllegalArgumentException("untranslated CAN-DO call");
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(Text list, character item)
   {
      if (list.isUnknown() || item.isUnknown())
         return new logical();
   
      return matchesList(list.getValue(), item.getValue());
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(Text list, String item)
   {
      if (list.isUnknown())
         return new logical();
   
      return matchesList(list.getValue(), item);
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(String list, character item)
   {
      if (item.isUnknown())
         return new logical();
   
      return matchesList(list, item.getValue());
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(String list, String item)
   {
      boolean result = false;
   
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(',');      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // checking item against every entry in the list until match is found
      int idx = 0;
      int i = 0;
      boolean negation = false;
      String piece = null;
      StringBuffer sb = new StringBuffer();
      char c = 0;
      Pattern p = null;
      Matcher m = null;
   
      for (idx = 0; idx < array.length; idx ++)
      {
         // isolate negation (!) case
         if (array[idx].length() > 0 && array[idx].charAt(0) == '!')
         {
            piece = array[idx].substring(1);
            negation = true;
         }
         else
         {
            piece = array[idx];
            negation = false;
         }
   
         // build a regex out of the list entry; only * is converted into .*
         sb.setLength(0);
         for (i = 0; i < piece.length(); i ++)
         {
            c = piece.charAt(i);
            if (c != '*')
               sb.append(fixupRegexChar(c));
            else
               sb.append(".*");
         }
         regex = sb.toString();
   
         // compile regex as case-insensitive and apply it to the item
         p = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);
         m = p.matcher(item);
         if (!m.matches())
            continue;
   
         // this is a matching entry from the list; produce the result
         result = !negation;
         break;
      }
   
      return new logical(result);
   }

   /**
    * Returns the largest instance (lexicographically) from the list.  This
    * list may not include <code>null</code> values.  Since an instance
    * is returned which is one of the references that was passed in, the
    * case-sensitivity of the input will be the same on output.
    *
    * @param    list
    *           The list of character instances to compare.
    *
    * @return   The largest instance or <code>unknown value</code> if any
    *           entries in the list are unknown or <code>null</code> if
    *           an empty list is passed.
    */
   public static <T extends Text> T maximum(T ... list)   
   {
      return (T) CompareOps.maximum(list);
   }

   /**
    * Returns the smallest instance (lexicographically) from the list.  This
    * list may not include <code>null</code> values.  Since an instance
    * is returned which is one of the references that was passed in, the
    * case-sensitivity of the input will be the same on output.
    *
    * @param    list
    *           The list of character instances to compare.
    *
    * @return   The smallest instance or <code>unknown value</code> if any
    *           entries in the list are unknown or <code>null</code> if
    *           an empty list is passed.
    */
   public static <T extends Text> T minimum(T ... list)   
   {
      return (T) CompareOps.minimum(list);
   }

   /**
    * Get the normalized form of the given string, based on the specified
    * Unicode normalization form:
    * The <code>format</code> parameter must evaluate to one of the following 
    * strings:
    * <ul>
    * <li> NFD - Canonical Decomposition
    * <li> NFC - Canonical Decomposition, followed by Canonical Composition
    * <li> NFKD - Compatibility Decomposition
    * <li> NFKC - Compatibility Decomposition, followed by Canonical Composition
    * <li> NONE - Returns the source string unchanged
    * </ul>
    * 
    * @param    text
    *           The text which needs to be normalized, a {@link character} or
    *           {@link longchar} instance.
    * @param    format
    *           The normalization format, one of the strings above.
    *
    * @return   The normalized form.
    */
   public static <T extends Text> T normalize(T text, String format)
   {
      return normalize(text, new character(format));
   }

   /**
    * Get the normalized form of the given string, based on the specified
    * Unicode normalization form:
    * The <code>format</code> parameter must evaluate to one of the following 
    * strings:
    * <ul>
    * <li> NFD - Canonical Decomposition
    * <li> NFC - Canonical Decomposition, followed by Canonical Composition
    * <li> NFKD - Compatibility Decomposition
    * <li> NFKC - Compatibility Decomposition, followed by Canonical Composition
    * <li> NONE - Returns the source string unchanged
    * </ul>
    * 
    * @param    text
    *           The text which needs to be normalized, a {@link character} or
    *           {@link longchar} instance.
    * @param    format
    *           The normalization format, one of the strings above.
    *
    * @return   The normalized form.
    */
   public static <T extends Text> T normalize(T text, character format)
   {
      // TODO: IMPLEMENT THIS
      return text;
   }
   
   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or -1 if the list is the
    *           <code>unknown value</code>.
    */
   public static int numEntriesOf(Text list, character delimit)
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return -1;
      }
      
      return numEntriesOf(list.getValue(), safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list or -1 if the list is the
    *           <code>unknown value</code>.
    */
   public static int numEntriesOf(Text list)
   {
      if (list.isUnknown())
      {
         return -1;
      }
      
      return numEntriesOf(list.getValue(), ',');
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list.
    */
   public static int numEntriesOf(String list, char delimit)
   {
      if (list.length() == 0)
      {
         return 0;
      }
      
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delimit);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      return array.length;
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list.
    */
   public static int numEntriesOf(String list)
   {
      return numEntriesOf(list, ',');
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(Text list)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue()));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(Text list, String delimit)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue(),
                                      safeDelimiter(delimit)));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(String list, character delimit)
   {
      if (delimit.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list,
                                      safeDelimiter(delimit.getValue())));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(String list, String delimit)
   {
      return new integer(numEntriesOf(list,
                                      safeDelimiter(delimit)));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(Text list, character delimit)
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue(),
                                      safeDelimiter(delimit.getValue())));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list.
    */
   public static integer numEntries(String list, char delimit)
   {
      return new integer(numEntriesOf(list, delimit));
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list.
    */
   public static integer numEntries(String list)
   {
      return new integer(numEntriesOf(list));
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static <T extends Text> T replaceAll(T    src,
                                               Text from,
                                               Text to)
   {
      if (src.isUnknown() || from.isUnknown() || to.isUnknown())
      {
         return (T) instantiateUnknown(src.getClass());
      }
      
      boolean cs = (src.isCaseSensitive()  || 
                    from.isCaseSensitive() || 
                    to.isCaseSensitive());
      
      return (T) replaceAllImpl(src.getClass(),
                                src.getValue(),
                                from.getValue(),
                                to.getValue(),
                                cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static <T extends Text> T replaceAll(T      src,
                                               Text   from,
                                               String to)
   {
      if (src.isUnknown() || from.isUnknown())
      {
         return (T) instantiateUnknown(src.getClass());
      }
      
      boolean cs = (src.isCaseSensitive() || from.isCaseSensitive());
      
      return (T) replaceAllImpl(src.getClass(), 
                                src.getValue(), 
                                from.getValue(),
                                to,
                                cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static <T extends Text> T replaceAll(T      src,
                                               String from,
                                               Text   to)
   {
      if (src.isUnknown() || to.isUnknown())
      {
         return (T) instantiateUnknown(src.getClass());
      }
      
      boolean cs = (src.isCaseSensitive() || to.isCaseSensitive());
      
      return (T) replaceAllImpl(src.getClass(),
                                src.getValue(),
                                from,
                                to.getValue(),
                                cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static character replaceAll(String src,
                                      Text   from,
                                      Text   to)
   {
      if (from.isUnknown() || to.isUnknown())
      {
         return new character();
      }
      
      boolean cs = (from.isCaseSensitive() || to.isCaseSensitive());
      
      return replaceAllImpl(character.class,
                            src,
                            from.getValue(),
                            to.getValue(),
                            cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>src</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>src</code> string is <code>unknown value</code>.
    */
   public static <T extends Text> T replaceAll(T      src,
                                               String from,
                                               String to)
   {
      if (src.isUnknown())
      {
         return (T) instantiateUnknown(src.getClass());
      }
      
      return (T) replaceAllImpl(src.getClass(), 
                                src.getValue(),
                                from,
                                to,
                                src.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>from</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>from</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, Text from, String to)
   {
      if (from.isUnknown())
      {
         return new character();
      }
      
      return replaceAllImpl(character.class,
                            src,
                            from.getValue(),
                            to,
                            from.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>to</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>to</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, String from, Text to)
   {
      if (to.isUnknown())
      {
         return new character();
      }
      
      return replaceAllImpl(character.class,
                            src,
                            from,
                            to.getValue(),
                            to.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are always case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>src</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, String from, String to)
   {
      return replaceAllImpl(character.class, src, from, to, false);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    * @param    cs
    *           Determines if a case-sensitive match should be performed AND
    *           the case-sensitivity of the returned instance.
    *
    * @return   The resulting string.
    */
   public static character replaceAll(String  src,
                                      String  from,
                                      String  to,
                                      boolean cs)
   {
      return replaceAllImpl(character.class, src, from, to, cs);
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T rightTrim(T input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) rightTrimImpl(input.getClass(), 
                               input.getValue(),
                               list.getValue(),
                               input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * the input parameter. If the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T rightTrim(T input, String list)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) rightTrimImpl(input.getClass(),
                               input.getValue(),
                               list,
                               input.isCaseSensitive());
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * right of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T rightTrim(T input)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) rightTrimImpl(input.getClass(),
                               input.getValue(),
                               null,
                               input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based case-insensitively.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return rightTrimImpl(character.class, input, list.getValue(), false);
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
   
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input, String list)
   {
      return rightTrimImpl(character.class, input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * right of an input string.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.    
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input)
   {
      return rightTrim(input, (String) null);
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String  input,
                                     String  list,
                                     boolean caseSens)
   {
      return rightTrimImpl(character.class, input, list, caseSens);
   }

   /**
    * Perform positional replacement of '&amp;n' specifiers in the format string
    * <code>fmt</code>, using the array of <code>character</code> and/or
    * <code>String</code> objects as the list of replacements.  The 1-based
    * index position of the '&amp;n' specifiers (where n and range from 1 to 
    * 9) is converted to a 0-based index into the <code>replacements</code>
    * array.  Instances of '&amp;&amp;' yield a single '&amp;' in the result.
    * Any specifiers that do not correspond to an object in the array are
    * replaced with an empty string. 
    *
    * @param    fmt
    *           The string specifying the text and format specifiers. 
    * @param    replacements
    *           The list of replacements.
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>fmt</code> string is <code>unknown value</code>.
    */
   public static <T extends Text> T substitute(T fmt, Object... replacements)
   throws ErrorConditionException
   {
      if (fmt.isUnknown())
         return (T) instantiateUnknown(fmt.getClass());
      
      return (T) substituteImpl(fmt.getClass(), fmt.getValue(), replacements);
   }

   /**
    * Perform positional replacement of '&amp;n' specifiers in the format string
    * <code>fmt</code>, using the array of <code>character</code> and/or
    * <code>String</code> objects as the list of replacements.  The 1-based
    * index position of the '&amp;n' specifiers (where n and range from 1 to 
    * 9) is converted to a 0-based index into the <code>replacements</code>
    * array.  Instances of '&amp;&amp;' yield a single '&amp;' in the result.
    * Any specifiers that do not correspond to an object in the array are
    * replaced with an empty string. 
    *
    * @param    fmt
    *           The string specifying the text and format specifiers. 
    * @param    replacements
    *           The list of replacements.
    *
    * @return   The resulting string.
    */
   public static character substitute(String fmt, Object... replacements)
   throws ErrorConditionException
   {
      return substituteImpl(character.class, fmt, replacements);
   }
   
   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              NumberType len,
                                              String     type)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len.intValue(),
                               type,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              NumberType len,
                                              character  type)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown() || type.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len.intValue(),
                               type.getValue(),
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              int        pos,
                                              NumberType len,
                                              String     type)
   {
      if (src.isUnknown() || len.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len.intValue(),
                               type,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              int        pos,
                                              NumberType len,
                                              character  type)
   {
      if (src.isUnknown() || len.isUnknown() || type.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len.intValue(),
                               type.getValue(),
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              int        len,
                                              String     type)
   {
      if (src.isUnknown() || pos.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len,
                               type,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              int        len,
                                              character  type)
   {
      if (src.isUnknown() || pos.isUnknown() || type.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len,
                               type.getValue(),
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T      src,
                                              int    pos,
                                              int    len,
                                              String type)
   {
      if (src.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len,
                               type,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T         src,
                                              int       pos,
                                              int       len,
                                              character type)
   {
      if (src.isUnknown() || type.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len,
                               type.getValue(),
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              NumberType len)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len.intValue(),
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              int        pos,
                                              NumberType len)
   {
      if (src.isUnknown() || len.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len.intValue(),
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T          src,
                                              NumberType pos,
                                              int        len)
   {
      if (src.isUnknown() || pos.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               len,
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T src, int pos, int len)
   {
      if (src.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               len,
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * extending for the rest of the string.   Operates on characters rather
    * than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T src, NumberType pos)
   {
      if (src.isUnknown() || pos.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos.intValue(),
                               -1,
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * extending for the rest of the string.  Operates on characters rather
    * than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static <T extends Text> T substring(T src, int pos)
   {
      if (src.isUnknown())
         return (T) instantiateUnknown(src.getClass());
      
      return (T) substringImpl(src.getClass(),
                               src.getValue(),
                               pos,
                               -1,
                               null,
                               src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index  and
    * extending for the rest of the string. Operates on characters rather than
    * bytes.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src, int pos)
   {
      return substringImpl(character.class, src, pos, -1, null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     NumberType pos,
                                     NumberType len)
   {
      if (pos.isUnknown() || len.isUnknown())
         return new character();
      
      return substringImpl(character.class, 
                           src,
                           pos.intValue(),
                           len.intValue(),
                           null,
                           false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     NumberType pos,
                                     int        len)
   {
      if (pos.isUnknown())
         return new character();
      
      return substringImpl(character.class, 
                           src,
                           pos.intValue(),
                           len,
                           null,
                           false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     int        pos,
                                     NumberType len)
   {
      if (len.isUnknown())
         return new character();
      
      return substringImpl(character.class, 
                           src,
                           pos,
                           len.intValue(),
                           null,
                           false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src, int pos, int len)
   {
      return substringImpl(character.class, src, pos, len, null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src,
                                     int    pos,
                                     int    len,
                                     String type)
   {
      return substringImpl(character.class, src, pos, len, type, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    * @param    caseSens
    *           Determines the case-sensitivity of the returned instance.
    *
    * @return   The resulting substring.
    */
   public static character substring(String  src,
                                     int     pos, 
                                     int     len,
                                     String  type,
                                     boolean caseSens)
   {
      return substringImpl(character.class, src, pos, len, type, caseSens);
   }
   
   /**
    * Returns a character instance that contains the lowercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * If the operand is case-sensitive, the result will be case-sensitive.
    *
    * @param    c
    *           The character instance to convert (cannot be
    *           <code>null</code>).
    *
    * @return   The lowercased character instance (which should match up
    *           with the Progress equivalents in the ASCII range) or
    *           <code>unknown value</code> if the input string is
    *           <code>unknown value</code>.
    */
   public static <T extends Text> T toLowerCase(T c)
   {
      if (c.isUnknown())
         return (T) instantiateUnknown(c.getClass());
      
      T res = (T) toLowerCaseImpl(c.getClass(), c.getValue());
      
      // honor case-sensitivity
      res.setCaseSensitive(c.isCaseSensitive());
      
      return res;
   }

   /**
    * Returns a character instance that contains the lowercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The lowercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static character toLowerCase(String str)
   {
      return toLowerCaseImpl(character.class, str);
   }

   /**
    * Returns a character instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * If the operand is case-sensitive, the result will be case-sensitive.
    *
    * @param    c
    *           The character instance to convert (cannot be
    *           <code>null</code>).
    *
    * @return   The uppercased character instance (which should match up
    *           with the Progress equivalents in the ASCII range) or
    *           <code>unknown value</code> if the input string is
    *           <code>unknown value</code>.
    */
   public static <T extends Text> T toUpperCase(T c)
   {
      if (c.isUnknown())
         return (T) instantiateUnknown(c.getClass());
      
      T res = (T) toUpperCaseImpl(c.getClass(), c.getValue());
      
      // honor case-sensitivity
      res.setCaseSensitive(c.isCaseSensitive());
      
      return res;
   }

   /**
    * Returns a character instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    *The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The uppercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static character toUpperCase(String str)
   {
      return toUpperCaseImpl(character.class, str);
   }

   /**
    * Returns a string instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    *The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    * @param    runtimeString
    *           <code>true</code> if the target string is a runtime string,
    *           <code>false</code> if it is a hardcoded string into a source
    *           file.
    *
    * @return   The uppercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static String toUpperCase(String str, boolean runtimeString)
   {
      str = str.toUpperCase();
   
      if (!runtimeString)
      {
         StringBuilder sb = new StringBuilder("");
         int end = str.length();
   
         for (int i = 0; i < end; i++)
         {
            char current = str.charAt(i);
            char next;
   
            switch (current)
            {
               case '\\':
                  i++;
                  next = str.charAt(i);
                  sb.append('\\');
   
                  // change back mistakenly uppercased special characters
                  // \t \r \n \b \f into a hardcoded string
                  switch (next)
                  {
                     case 'T':
                        sb.append('t');
                        break;
                     case 'R':
                        sb.append('r');
                        break;
                     case 'N':
                        sb.append('n');
                        break;
                     case 'B':
                        sb.append('b');
                        break;
                     case 'F':
                        sb.append('f');
                        break;
                     default:
                        sb.append(next);
                  }
                  break;
               default:
                  sb.append(current);
            }
         }
   
         str = sb.toString();
      }
   
      return str;
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T trim(T input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) trimImpl(input.getClass(),
                          input.getValue(),
                          list.getValue(),
                          input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter. Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T trim(T input, String list)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) trimImpl(input.getClass(),
                          input.getValue(),
                          list,
                          input.isCaseSensitive());
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left and right of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static <T extends Text> T trim(T input)
   {
      if (input.isUnknown())
      {
         return (T) instantiateUnknown(input.getClass());
      }
      
      return (T) trimImpl(input.getClass(), 
                          input.getValue(),
                          null,
                          input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed case-insensitively and the returned
    * result will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return trimImpl(character.class, input, list.getValue(), false);
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string. If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, String list)
   {
      return trimImpl(character.class, input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left and right of an input string.  Trim operations are performed
    * case-insensitively.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.    
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input)
   {
      return trimImpl(character.class, input, (String) null, false);
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string. If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, String list, boolean caseSens)
   {
      // trim left
      String result = trim(input, list, true, caseSens);
      
      // trim right
      return new character(trim(result, list, false, caseSens), caseSens);
   }

   /**
    * Generate an error for a substring or overlay index that is negative.
    */
   static void genNegativeIndexError()
   {
      String err = "Starting position for SUBSTRING, OVERLAY, etc. must" +
                   " be 1 or greater";
      ErrorManager.recordOrThrowError(82, err);
   }

   /**
    * Generate an error for a value out of range.
    *
    * @param    idx
    *           The index that is out of range.
    * @param    list
    *           The list being indexed.
    */
   static void genOutsideRangeError(int idx, String list)
   {
      String spec = "Entry %d is outside the range of list %s";
      String err  = String.format(spec, idx, list);
      
      ErrorManager.recordOrThrowError(560, err);
   }

   /**
    * Converts a single character into a regular expression, including
    * escaping any of the special characters that have other meaning in
    * regular expressions.
    *
    * @param    ch
    *           The exact character on which to match.
    *
    * @return   The resulting regular expression.
    */
   static String fixupRegexChar(char ch)
   {
      StringBuffer regex = new StringBuffer();
      
      if (ch == '(' || ch == ')' || ch == '.' || ch == '*' || ch == '[' ||
          ch == ']' || ch == '^' || ch == '{' || ch == '}' || ch == '?' ||
          ch == '+' || ch == '?' || ch == '$' || ch == '|' || ch == '\\')
      {
         regex.append('\\').append(ch);
      }
      else
      {
         regex.append(ch);
      }
      
      return regex.toString();
   }

   /**
    * Given a delimiter string, extract the first character and retrun it.  If
    * the string is empty, return a space character.
    * 
    * @param   delimiter
    *          Delimiter string
    * 
    * @return  Delimiter character as described above.
    */
   static char safeDelimiter(String delimiter)
   {
      return (delimiter.length() == 0 ? ' ' : delimiter.charAt(0));
   }

   /**
    * Convert the compare operator's string representation to its numeric 
    * representation, to be used by the {@link #compare(Text, int, Text, int)}
    * method.
    *  
    * @param    compare
    *           The operator's string representation.
    *           
    * @return   The numeric representation.
    */
   private static int getCompareOperator(character compare)
   {
      // TODO: IMPLEMENT THIS
      return -1;
   }

   /**
    * Convert the strength's string representation to its numeric 
    * representation, to be used by the {@link #compare(Text, int, Text, int)}
    * method.
    *  
    * @param    strength
    *           The strength's string representation.
    *           
    * @return   The numeric representation.
    */
   private static int getStrength(character strength)
   {
      // TODO: IMPLEMENT THIS
      return -1;
   }
   
   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    ii
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   private static <T extends Text> T entryImpl(Class<T> cls,
                                               double ii,
                                               String list,
                                               char delimit)
   throws ErrorConditionException
   {
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delimit);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // convert 1 based index to 0 based
      int i   = (int) ii;
      int idx = i - 1;
      
      if (idx < 0 || idx >= array.length)
      {
         genOutsideRangeError(i, list);
         
         // this is the safest thing to return (it should be undone anyway)
         return instantiate(cls, "");
      }
      
      return instantiate(cls, array[idx]);
   }

   /**
    * Create a new instance of the given {@link Text} subclass and set it to 
    * unknown.
    * 
    * @param    cls
    *           A {@link Text} subclass for which an unknown instance is 
    *           needed.
    *           
    * @return   See above.
    */
   private static <T extends Text> T instantiateUnknown(Class<T> cls)
   {
      try
      {
         return cls.newInstance();
      }
      catch (Exception e)
      {
         // TODO: should this be ignored ?
         throw new StopConditionException(e.getMessage());
      }
   }

   /**
    * Create a new instance of the given {@link Text} subclass and set its
    * value.  If the <code>val</code> is null, the new instance will be set
    * to unknown.
    * 
    * @param    cls
    *           A {@link Text} subclass for which a new instance is needed.
    * @param    val
    *           The value to set for the new instance.
    *           
    * @return   See above.
    */
   private static <T extends Text> T instantiate(Class<T> cls, Object val)
   {
      if (val == null)
      {
         return instantiateUnknown(cls);
      }
   
      try
      {
         return cls.getConstructor(val.getClass()).newInstance(val);
      }
      catch (Exception e)
      {
         // TODO: should this be ignored ?
         throw new StopConditionException(e.getMessage());
      }
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   private static <T extends Text> T leftTrimImpl(Class<T> cls,
                                                  String  input,
                                                  String  list,
                                                  boolean caseSens)
   {
      T res = (T) instantiate(cls, trim(input, list, true, caseSens));
      res.setCaseSensitive(caseSens);
      
      return res;
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    * @param    cs
    *           Determines if a case-sensitive match should be performed AND
    *           the case-sensitivity of the returned instance.
    *
    * @return   The resulting string.
    */
   private static <T extends Text> T replaceAllImpl(Class<T> cls,
                                                    String   src,
                                                    String   from,
                                                    String   to,
                                                    boolean  cs)
   {
      StringBuffer sb = new StringBuffer();
      
      int len  = src.length();
      int flen = from.length(); 
      
      for (int i = 0; i < len; i++)
      {
         if (src.regionMatches(!cs, i, from, 0, flen)) 
         {
            // make a replacement
            sb.append(to);
            
            // move past the end of the replaced text (take into account that
            // as the loop iterates, i will be incremented again)
            i += (flen - 1);
         }
         else
         {
            // no match, put the char into the output
            sb.append(src.charAt(i));
         }
      }
      
      T res = instantiate(cls, sb.toString());
      res.setCaseSensitive(cs);
      
      return res;
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   private static <T extends Text> T rightTrimImpl(Class<T> cls,
                                                   String   input,
                                                   String   list,
                                                   boolean  caseSens)
   {
      T res = instantiate(cls, trim(input, list, false, caseSens));
      res.setCaseSensitive(caseSens);
      
      return res;
   }

   /**
    * Perform positional replacement of '&amp;n' specifiers in the format string
    * <code>fmt</code>, using the array of <code>character</code> and/or
    * <code>String</code> objects as the list of replacements.  The 1-based
    * index position of the '&amp;n' specifiers (where n and range from 1 to 
    * 9) is converted to a 0-based index into the <code>replacements</code>
    * array.  Instances of '&amp;&amp;' yield a single '&amp;' in the result.
    * Any specifiers that do not correspond to an object in the array are
    * replaced with an empty string. 
    *
    * @param    fmt
    *           The string specifying the text and format specifiers. 
    * @param    replacements
    *           The list of replacements.
    *
    * @return   The resulting string.
    */
   private static <T extends Text> T substituteImpl(Class<T> cls,
                                                    String fmt,
                                                    Object... replacements)
   throws ErrorConditionException
   {
      StringBuffer sb = new StringBuffer();
      
      int len  = fmt.length();
      
      for (int i = 0; i < len; i++)
      {
         char c = fmt.charAt(i);
         
         if (c == '&') 
         {
            // ignore this char and look at the next one
            i++;
            c = fmt.charAt(i);
            
            if (c >= '0' && c <= '9')
            {
               // convert this character into a 0-based index into the array
               int idx = Character.digit(c, 10) - 1;
               
               // check if this is a valid replacement index (note that 0 is
               // always ignored/invalid), if not the result is an empty
               // string since we don't make any replacement in that case
               if (idx >= 0 && idx < replacements.length) 
               {
                  Object obj = replacements[idx];
                  
                  if (obj != null)
                  {
                     // make a replacement if it is an object we recognize
                     if (obj instanceof String)
                     {
                        sb.append((String) obj);   
                     }
                     else if (obj instanceof character)
                     {
                        character ch = (character) obj;
                        sb.append(ch.isUnknown() ? "?" : ch.getValue());
                     }
                     else if (obj instanceof BaseDataType)
                     {
                        BaseDataType bdt = (BaseDataType) obj;
                        sb.append(bdt.toStringExport());
                     }
                  }
               }
            }            
            else if (c == '&')
            {
               sb.append(c);
            }
            else
            {
               String spec = "Invalid character \'%c\' following \'&\'" +
                             " in SUBSTITUTE";
               ErrorManager.recordOrThrowError(2920, String.format(spec, c));
               
               // this is the safest thing to return (it should be undone anyway)
               return instantiate(cls, "");
            }
         }
         else
         {
            // no match, put the char into the output
            sb.append(c);
         }
      }
      
      return instantiate(cls, sb.toString());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    * @param    caseSens
    *           Determines the case-sensitivity of the returned instance.
    *
    * @return   The resulting substring.
    */
   private static <T extends Text> T substringImpl(Class<T> cls,
                                                   String  src,
                                                   int     pos, 
                                                   int     len,
                                                   String  type,
                                                   boolean caseSens)
   {
      String result = null;
      
      // convert the Progress 1-based index into Java's 0-based index
      int start = pos - 1;
      
      if (pos < 1)
      {
         genNegativeIndexError();
         return instantiate(cls, "");
      }
      
      if (type != null && "raw".equalsIgnoreCase(type))
      {
         // raw mode (operates on bytes)
         byte[] buffer = src.getBytes(); 
         
         if (start >= buffer.length)
         {
            // requesting data past the end of the string yields the empty
            // string
            result = "";
         }
         else
         {
            int end = start + len;
            
            if (len == -1 || end >= buffer.length)
            {
               len = buffer.length - start;
            }
            
            result = new String(buffer, start, len);         
         }
      }
      else
      {      
         // character mode
         
         if (start >= src.length())
         {
            // requesting data past the end of the string yields the empty
            // string
            result = "";
         }
         else
         {
            int end = start + len;
            
            if (len == -1 || end >= src.length())
            {
               // if the requested data is more than the rest of the string
               // or if the rest of the string was explicitly requested via
               // -1
               result = src.substring(start);
            }
            else
            {
               // only return a portion
               result = src.substring(start, end);
            }            
         }         
      }
      
      T res = instantiate(cls, result);
      res.setCaseSensitive(caseSens);
      
      return res;
   }

   /**
    * Returns a character instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    *The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The uppercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   private static <T extends Text> T toUpperCaseImpl(Class<T> cls, String str)
   {
      return instantiate(cls, str.toUpperCase());
   }

   /**
    * Returns a character instance that contains the lowercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The lowercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   private static <T extends Text> T toLowerCaseImpl(Class<T> cls, String str)
   {
      return instantiate(cls, str.toLowerCase());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string. If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   private static <T extends Text> T trimImpl(Class<T> cls, 
                                              String   input,
                                              String   list,
                                              boolean  caseSens)
   {
      // trim left
      String result = trim(input, list, true, caseSens);
      
      // trim right
      T res = instantiate(cls, trim(result, list, false, caseSens));
      res.setCaseSensitive(caseSens);
      
      return res;
   }

   /**
    * Trims the list of contiguous trim characters from the left or right of
    * an input string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    left
    *           Specifies if the trim should occur on the left or right
    *           side of the string.
    * @param    caseSens
    *           If <code>true</code>, a case-sensitive trim is performed.
    *
    * @return   The trimmed result.
    */
   private static String trim(String  input,
                              String  list,
                              boolean left,
                              boolean caseSens)
   {
      // implement our default list
      if (list == null)
      {
         list = " \t\n\r";
      }
      
      int len  = input.length();
      int llen = list.length();
      int i;
      
      for (i = left ? 0 : len - 1; left ? i < len : i >= 0;)
      {
         boolean found = false;
         char    c     = input.charAt(i);
         char    cu    = Character.toUpperCase(c);
         
         for (int j = 0; j < llen; j++)
         {
            char ch  = list.charAt(j);
            char chu = Character.toUpperCase(ch);
            
            if ((caseSens && ch == c) || (!caseSens && chu == cu))  
            {
               found = true;
               break;
            }
         }
         
         if (!found)
         {
            break;
         }
         
         if (left)
         {
            i++;
         }
         else
         {
            i--;
         }
      }
      
      return left ? input.substring(i) : input.substring(0, i + 1);
   }

   /**
    * Returns the length of the given <code>character</code> instance, using the type parameter
    * to determine the measurement.
    * This method checks the <code>type</code> argument and decide to what method to delegate
    * the actual implementation depending on the 2nd parameter.
    *
    * @param    c
    *           The instance to test.
    * @param    type
    *           The measurement mode.
    *
    * @return   The length of the given text
    */
   private static integer length(String c, String type)
   {
      if (type == null)
      {
         return new integer();
      }
      else if ("raw".equalsIgnoreCase(type))
      {
         return TextOps.byteLength(c);
      }
      else if ("character".equalsIgnoreCase(type))
      {
         return TextOps.length(c);
      }
      else
      {
         // "column" is not supported.
         /* 4GL will always return a value if both arguments are not unknown (?)
            I was not able to detect and there is no detail about the default 
            mode when an invalid mode is detected at runtime */
         return TextOps.length(c);
      }
   }
}
