/*
** Module   : UnimplementedFeature.java
** Abstract : helper class to cleanly log runtime usage of code that is unimplemented or partial
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 GES 20141029 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.logging.*;
import java.util.*;

/**
 * Helper class to cleanly log runtime usage of code that is unimplemented or partial.
 * Each log entry will contain a description as well as a stack trace identifying the call
 * path that tried to use the unimplemented feature.
 * <p>
 * <b>Usage Guidelines</b>
 * <p>
 * <ul>
 *   <li> {@link #missing} should be used for methods/features which have no substantial
 *        implementation provided.
 *   <li> {@link #partial} should be used for methods/features which have some implementation
 *        but it is not complete.  If possible, place the call to <code>partial</code> inside
 *        code that only executes when the missing portions of the implementation would be
 *        executed.
 *   <li> {@link #todo} should be used for anything that needs further review, investigation
 *        or modification.  If possible, place the call to <code>todo</code> inside code
 *        that only executes when the associated condition presents itself.
 *   <li> There are forms of each method that take and return a token.  Use these forms if
 *        the logging should only occur the first time the code is executed. This is useful
 *        for features that get called with frequency such that the log file would get filled
 *        with "noise" from excess logging.
 *   <li> When using tokens, they should generally be unique to the specific calling location
 *        otherwise you will have log entries suppressed from unrelated calls to this class.
 *   <li> When the feature is actually implemented, the logging call to this class would 
 *        naturally be remvoed.  <b>Don't forget to delete any associated token data members
 *        at that time!</b> 
 * </ul>
 * <p>
 * Example of using tokens:
 * <p>
 * <pre>
 * private static Object token = null;
 * 
 * public void someMethod()
 * {
 *    if (someCondition)
 *    {
 *       // the first time this is called, the token is null and the log entry will be made
 *       // but subsequent calls will pass the token returned from the first call and the
 *       // logging will be suppressed
 *       token = UnimplementedFeature.todo("This is a description of the TODO.", token);
 *    }
 * }
 * </pre>
 */
public class UnimplementedFeature
{
   /** Logger */
   private static final Logger logger = LogHelper.getLogger(UnimplementedFeature.class);
   
   /** Set of all tokens provided to callers. */ 
   private static Set tokens = new HashSet(); 
   
   /**
    * Log the given message along with a given stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support is completely missing. The
    * logging is unconditionally done (there is no logging level checked) with the level set
    * to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the missing feature will
    * already be readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    */
   public static void missing(String msg)
   {
      log("MISSING", msg, new RuntimeException());
   }
   
   /**
    * Conditionally log the given message if the caller's provided token is <code>null</code>
    * or if it was not a valid token.  Subsequent calls to this method will silently return
    * without logging if the given token already exists in the set of valid tokens. Any log entry
    * will have the given text along with the a stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support is completely missing. The
    * logging is unconditionally done (there is no logging level checked) with the level set
    * to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the missing feature will
    * already be readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    * @param    token
    *           A previously provided token value that uniquely distinguishes the caller
    *           or <code>null</code> if this is the first call.
    *
    * @return   The given token if it is a valid previously provided token.  Otherwise a newly
    *           created token will be associated with the caller and returned.
    */
   public static Object missing(String msg, Object token)
   {
      if (token == null || !tokens.contains(token))
      {
         token = new Object();
         tokens.add(token);
         missing(msg);
      }
      
      return token;
   }
   
   /**
    * Log the given message along with the a stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support is only partially implemented.
    * The logging is unconditionally done (there is no logging level checked) with the level
    * set to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the partial feature will
    * already be readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    */
   public static void partial(String msg)
   {
      log("PARTIAL", msg, new RuntimeException());
   }
   
   /**
    * Conditionally log the given message if the caller's provided token is <code>null</code>
    * or if it was not a valid token.  Subsequent calls to this method will silently return
    * without logging if the given token already exists in the set of valid tokens. Any log entry
    * will have the given text along with the a stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support is only partially implemented.
    * The logging is unconditionally done (there is no logging level checked) with the level
    * set to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the partial feature will
    * already be readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    * @param    token
    *           A previously provided token value that uniquely distinguishes the caller
    *           or <code>null</code> if this is the first call.
    *
    * @return   The given token if it is a valid previously provided token.  Otherwise a newly
    *           created token will be associated with the caller and returned.
    */
   public static Object partial(String msg, Object token)
   {
      if (token == null || !tokens.contains(token))
      {
         token = new Object();
         tokens.add(token);
         partial(msg);
      }
      
      return token;
   }
   
   /**
    * Log the given message along with the a stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support has a TODO remaining. The
    * logging is unconditionally done (there is no logging level checked) with the level set
    * to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the TODO will already be
    * readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    */
   public static void todo(String msg)
   {
      log("TODO", msg, new RuntimeException());
   }
   
   /**
    * Conditionally log the given message if the caller's provided token is <code>null</code>
    * or if it was not a valid token.  Subsequent calls to this method will silently return
    * without logging if the given token already exists in the set of valid tokens. Any log entry
    * will have the given text along with the a stack trace and a prefix that indicates that
    * a feature was used at runtime for which the runtime support has a TODO remaining. The
    * logging is unconditionally done (there is no logging level checked) with the level set
    * to <code>SEVERE</code>.
    * <p>
    * Since a call stack will be created and logged, the location of the TODO will already be
    * readily apparent.
    *
    * @param    msg
    *           A description of the missing feature.
    * @param    token
    *           A previously provided token value that uniquely distinguishes the caller
    *           or <code>null</code> if this is the first call.
    *
    * @return   The given token if it is a valid previously provided token.  Otherwise a newly
    *           created token will be associated with the caller and returned.
    */
   public static Object todo(String msg, Object token)
   {
      if (token == null || !tokens.contains(token))
      {
         token = new Object();
         tokens.add(token);
         todo(msg);
      }
      
      return token;
   }
   
   /**
    * Logging helper method that logs a <code>SEVERE</code> message.
    *
    * @param    code
    *           Describes the type of unimplemented feature.
    * @param    msg
    *           The caller-defined message.
    * @param    exc
    *           Exception that contains the call stack to include in the log entry.
    */
   private static void log(String code, String msg, RuntimeException exc)
   {
      String txt = String.format("%s: %s", code, msg);
      
      // the code is passed here but not honored, so we add it explicitly above
      logger.logp(Level.SEVERE, "UnimplementedFeature", code, LogHelper.generate(txt), exc);
   }
}
